\ 128-bit AES Decryption
\ Author: SevenW from sevenwatt.com
\ Date  : 2017-Feb-23
\
\ Description:
\ AES128 block cipher decryption
\ Implementation is optimized on (low) memory usage.
\ For AES-CTR and AES-CMAC as used in LoraWAN, this module is NOT needed.
\ A testmodule is provided in the file aes128invtest.fs
\
\ Requires : aes128.fs
\
\ Usage    : ( c-addr key ) -aes
\ With     : c-addr : input data in a 16-byte buffer
\            key    : the orignal encryption key in 16-bytes
\ Output   : Decryption is in-situ so the 16-byte input data buffer contains the decrypted output.
\
\ Adapted for DX-Forth and accelerated by Pavel Zampach, 2022

marker AES128DE

hex
create S.BOX.INV
52 c, 09 c, 6A c, D5 c, 30 c, 36 c, A5 c, 38 c, BF c, 40 c, A3 c, 9E c, 81 c, F3 c, D7 c, FB c,
7C c, E3 c, 39 c, 82 c, 9B c, 2F c, FF c, 87 c, 34 c, 8E c, 43 c, 44 c, C4 c, DE c, E9 c, CB c,
54 c, 7B c, 94 c, 32 c, A6 c, C2 c, 23 c, 3D c, EE c, 4C c, 95 c, 0B c, 42 c, FA c, C3 c, 4E c,
08 c, 2E c, A1 c, 66 c, 28 c, D9 c, 24 c, B2 c, 76 c, 5B c, A2 c, 49 c, 6D c, 8B c, D1 c, 25 c,
72 c, F8 c, F6 c, 64 c, 86 c, 68 c, 98 c, 16 c, D4 c, A4 c, 5C c, CC c, 5D c, 65 c, B6 c, 92 c,
6C c, 70 c, 48 c, 50 c, FD c, ED c, B9 c, DA c, 5E c, 15 c, 46 c, 57 c, A7 c, 8D c, 9D c, 84 c,
90 c, D8 c, AB c, 00 c, 8C c, BC c, D3 c, 0A c, F7 c, E4 c, 58 c, 05 c, B8 c, B3 c, 45 c, 06 c,
D0 c, 2C c, 1E c, 8F c, CA c, 3F c, 0F c, 02 c, C1 c, AF c, BD c, 03 c, 01 c, 13 c, 8A c, 6B c,
3A c, 91 c, 11 c, 41 c, 4F c, 67 c, DC c, EA c, 97 c, F2 c, CF c, CE c, F0 c, B4 c, E6 c, 73 c,
96 c, AC c, 74 c, 22 c, E7 c, AD c, 35 c, 85 c, E2 c, F9 c, 37 c, E8 c, 1C c, 75 c, DF c, 6E c,
47 c, F1 c, 1A c, 71 c, 1D c, 29 c, C5 c, 89 c, 6F c, B7 c, 62 c, 0E c, AA c, 18 c, BE c, 1B c,
FC c, 56 c, 3E c, 4B c, C6 c, D2 c, 79 c, 20 c, 9A c, DB c, C0 c, FE c, 78 c, CD c, 5A c, F4 c,
1F c, DD c, A8 c, 33 c, 88 c, 07 c, C7 c, 31 c, B1 c, 12 c, 10 c, 59 c, 27 c, 80 c, EC c, 5F c,
60 c, 51 c, 7F c, A9 c, 19 c, B5 c, 4A c, 0D c, 2D c, E5 c, 7A c, 9F c, 93 c, C9 c, 9C c, EF c,
A0 c, E0 c, 3B c, 4D c, AE c, 2A c, F5 c, B0 c, C8 c, EB c, BB c, 3C c, 83 c, 53 c, 99 c, 61 c,
17 c, 2B c, 04 c, 7E c, BA c, 77 c, D6 c, 26 c, E1 c, 69 c, 14 c, 63 c, 55 c, 21 c, 0C c, 7D c,
decimal

code 3cxor   ( b0 b1 b2 b3 -- b4 )
	H	pop
	L A	mov
	H	pop
	L	xra
	H	pop
	L	xra
	H	pop
	L	xra
	apush	jmp
end-code

\ substitute scratch with bytes from S.BOX.INV
: ~S.B-ALL@   ( -- )
  scratch aeslen bounds do
    i c@ s.box.inv + c@ i c!
  loop
;

create ~SC-TAB
 0 c,  1 c,  2 c,  3 c,  5 c,  6 c,  7 c,  4 c,
10 c, 11 c,  8 c,  9 c, 15 c, 12 c, 13 c, 14 c,

: ~SH-BYTES   ( -- )
  aeslen 0 do
    scratch i + c@
    ~sc-tab i + c@ sc-h + c!
  loop
  sc-h scratch aeslen cmove
;

\ galios field multiplication
code GMUL   ( a-addr b -- p )
	D	pop	\ b -> E
	H	pop	\ a -> L
	M L	mov
	B	push	\ save BC
	8 B	mvi	\ loop counter
	0 H	mvi	\ p -> H
1 $:	E 0	bit	\ bit 0 reset -> jump
	2 $	jrz
	H A	mov	\ p = p xor a
	L	xra
	A H	mov
2 $:	L A	mov	\ gfmod(a)
	A	add
	L 7	bit
	3 $	jrz
	$1b	xri
3 $:	A L 	mov
	E	srlr	\ b = b/2
	1 $	djnz
	B	pop	\ restore BC
	H A	mov
	apush	jmp
end-code

m1 1+ constant m1_1
m1 2+ constant m1_2
m1 3+ constant m1_3

\ multiply column with vector [ 14, 9, 13, 11 ] for decryption
: ~MIX1COL   ( c -- )
  4 0 do				\ copy column to m1
    dup i 4* + scratch + c@ m1 i + c!
  loop

  m1   14 gmul m1_3 9 gmul m1_2 13 gmul m1_1 11 gmul 3cxor m2    c!
  m1_1 14 gmul m1   9 gmul m1_3 13 gmul m1_2 11 gmul 3cxor m2 1+ c!
  m1_2 14 gmul m1_1 9 gmul m1   13 gmul m1_3 11 gmul 3cxor m2 2+ c!
  m1_3 14 gmul m1_2 9 gmul m1_1 13 gmul m1   11 gmul 3cxor m2 3+ c!

  4 0 do				\ copy m2 to column
    m2 i + c@ over i 4* + scratch + c!
  loop   drop
;

: ~MIX-COL   ( -- )
  4 0 do
    i ~mix1col
  loop
;

16 11 * buffer: ROUND.KEYS

: EXPAND-KEY   ( key -- )
  key-in
  round.key round.keys aeslen cmove
  11 1 do
    i round-key
    round.key round.keys i 4 lshift + aeslen cmove
  loop
;

: ROUND-KEY-LUT   ( round -- )
  1- 4 lshift round.keys + round.key aeslen cmove
;

: ~ONE-ROUND   ( round -- )
  dup 10 <> if ~mix-col then
  ~sh-bytes
  ~s.b-all@
  round-key-lut
  round-key+
;

\ aes128 decrypt block
\ input block is deciphered in-situ
: ~AES   ( c-addr key -- )
  expand-key
  dup >aes
  11 round-key-lut
  round-key+
  1 10 do
    i ~one-round
  -1 +loop
  aes>
;

\ alias
aka ~aes -AES
