\ 128-bit AES
\ Author: SevenW from sevenwatt.com
\ Date  : 2017-Feb-21
\
\ Description:
\ AES128 block cipher encryption
\ Implementation is optimized on (low) memory usage.
\ For AES-CTR and AES-CMAC as used in LoraWAN, only AES-encryption is needed.
\ Decryption is suported in the separate file: aes128inv.fs
\ A testmodule is provided in the file aes128test.fs
\
\ Usage    : ( c-addr key ) +aes
\ With     : c-addr : input data in a 16-byte buffer
\            key    : the encryption key in 16-bytes
\ Output   : Encryption is in-situ so the 16-byte input data buffer contains the encrypted output.
\
\ Adapted for DX-Forth, fixed and accelerated by Pavel Zampach, 2022

marker AES128EN

hex
create S.BOX
63 c, 7C c, 77 c, 7B c, F2 c, 6B c, 6F c, C5 c, 30 c, 01 c, 67 c, 2B c, FE c, D7 c, AB c, 76 c,
CA c, 82 c, C9 c, 7D c, FA c, 59 c, 47 c, F0 c, AD c, D4 c, A2 c, AF c, 9C c, A4 c, 72 c, C0 c,
B7 c, FD c, 93 c, 26 c, 36 c, 3F c, F7 c, CC c, 34 c, A5 c, E5 c, F1 c, 71 c, D8 c, 31 c, 15 c,
04 c, C7 c, 23 c, C3 c, 18 c, 96 c, 05 c, 9A c, 07 c, 12 c, 80 c, E2 c, EB c, 27 c, B2 c, 75 c,
09 c, 83 c, 2C c, 1A c, 1B c, 6E c, 5A c, A0 c, 52 c, 3B c, D6 c, B3 c, 29 c, E3 c, 2F c, 84 c,
53 c, D1 c, 00 c, ED c, 20 c, FC c, B1 c, 5B c, 6A c, CB c, BE c, 39 c, 4A c, 4C c, 58 c, CF c,
D0 c, EF c, AA c, FB c, 43 c, 4D c, 33 c, 85 c, 45 c, F9 c, 02 c, 7F c, 50 c, 3C c, 9F c, A8 c,
51 c, A3 c, 40 c, 8F c, 92 c, 9D c, 38 c, F5 c, BC c, B6 c, DA c, 21 c, 10 c, FF c, F3 c, D2 c,
CD c, 0C c, 13 c, EC c, 5F c, 97 c, 44 c, 17 c, C4 c, A7 c, 7E c, 3D c, 64 c, 5D c, 19 c, 73 c,
60 c, 81 c, 4F c, DC c, 22 c, 2A c, 90 c, 88 c, 46 c, EE c, B8 c, 14 c, DE c, 5E c, 0B c, DB c,
E0 c, 32 c, 3A c, 0A c, 49 c, 06 c, 24 c, 5C c, C2 c, D3 c, AC c, 62 c, 91 c, 95 c, E4 c, 79 c,
E7 c, C8 c, 37 c, 6D c, 8D c, D5 c, 4E c, A9 c, 6C c, 56 c, F4 c, EA c, 65 c, 7A c, AE c, 08 c,
BA c, 78 c, 25 c, 2E c, 1C c, A6 c, B4 c, C6 c, E8 c, DD c, 74 c, 1F c, 4B c, BD c, 8B c, 8A c,
70 c, 3E c, B5 c, 66 c, 48 c, 03 c, F6 c, 0E c, 61 c, 35 c, 57 c, B9 c, 86 c, C1 c, 1D c, 9E c,
E1 c, F8 c, 98 c, 11 c, 69 c, D9 c, 8E c, 94 c, 9B c, 1E c, 87 c, E9 c, CE c, 55 c, 28 c, DF c,
8C c, A1 c, 89 c, 0D c, BF c, E6 c, 42 c, 68 c, 41 c, 99 c, 2D c, 0F c, B0 c, 54 c, BB c, 16 c,
decimal

16 constant AESLEN
aeslen buffer: SCRATCH

code 3+   ( n -- n+3 )
	H	pop
	H	inx
	H	inx
	H	inx
		1push
end-code

code 4*   ( n -- n*4 )
	H	pop
	H	dad
	H	dad
		1push
end-code

code 4cxor   ( b0 b1 b2 b3 b4 -- b5 )
	H	pop
	L A	mov
	H	pop
	L	xra
	H	pop
	L	xra
	H	pop
	L	xra
	H	pop
	L	xra
	apush	jmp
end-code

\ lookup byte in S.BOX
: S.B@  ( b -- b )
  s.box + c@
;

\ substitute scratch with bytes from S.BOX
: S.B-ALL@   ( -- )
  scratch aeslen bounds do
    i c@ s.b@ i c!
  loop
;

aeslen buffer: SC-H			\ help buffer for shift bytes
create SC-TAB
 0 c,  1 c,  2 c,  3 c,  7 c,  4 c,  5 c,  6 c,
10 c, 11 c,  8 c,  9 c, 13 c, 14 c, 15 c, 12 c,

: SH-BYTES   ( -- )
  aeslen 0 do
    scratch i + c@
    sc-tab i + c@ sc-h + c!
  loop
  sc-h scratch aeslen cmove
;

4 buffer: M1
4 buffer: M2

\ multiply column with vector [ 2, 3, 1, 1 ] for encryption
\ gfmod = galios field multiply modulus
code M1M2   ( r c -- )
	D	pop	\ c->DE
	H	pop	\ r->HL
	B	push	\ save BC
	L C	mov	\ r->BC
	H B	mov
	H	dad	\ 4*r->HL
	H	dad
	D	dad	\ 4*r+c->HL
      scratch D	lxi
	D	dad	\ 4*r+c+scratch->HL
	M E	mov	\ b1->E
	E A	mov	\ gfmod(b1)->A
	A	add
	E 7	bit
	1 $	jrz
	$1b	xri
1 $:	m2 H	lxi	\ b2->(m2+r)
	B	dad
	A M	mov
	m1 H	lxi	\ b1->(m1+r)
	B	dad
	E M	mov
	B	pop	\ restore BC
		next
end-code

\ ##### fixed
: MIXING   ( i -- )
  scratch + >r
  m2 c@ m1 1+ c@ m2 1+ c@ m1 2+ c@ m1 3+ c@ 4cxor r@      c!
  m1 c@ m2 1+ c@ m1 2+ c@ m2 2+ c@ m1 3+ c@ 4cxor r@  4 + c!
  m1 c@ m1 1+ c@ m2 2+ c@ m1 3+ c@ m2 3+ c@ 4cxor r@  8 + c!
  m1 c@ m2    c@ m1 1+ c@ m1 2+ c@ m2 3+ c@ 4cxor r> 12 + c!
;

\ mix columns
: MIX-COL   ( -- )
  4 0 do
    4 0 do
      i j m1m2
    loop
    i mixing				\ ##### fixed
  loop
;

aeslen buffer: ROUND.KEY
4 buffer: RK-VAL

: RK-INIT   ( -- )
  round.key 12 + rk-val 4 cmove
;

: RK-ROTSUB   ( -- )
  rk-val dup c@ s.b@ swap
  3 0 do
    dup i + dup 1+ c@ s.b@ swap c!
  loop
  3+ c! ;

\ calculate rcon
: RCON   ( round -- rcon )
  1 swap				\ rcon round
  begin
    dup 1 <>
  while
    swap dup $80 and swap 2* swap if $1b xor then
    swap 1-
  repeat   drop
;

: XOR-RCON   ( round -- )
  rk-val dup c@ rot rcon xor swap c!	\ only first byte xor
;

: RK-CALC   ( ii io -- )
  4* over + round.key +			\ ii rki
  swap rk-val +				\ rki rkvali
  over c@ over c@ xor			\ rki rkvali val
  tuck swap c! swap c!
;

: RK-UPDATE   ( -- )
  4 0 do
    4 0 do
      i j rk-calc
    loop
  loop
;

\ Rijndael (incremental) key expansion
: ROUND-KEY   ( round -- )
  rk-init rk-rotsub xor-rcon rk-update
;

: RK+CALC   ( r c -- )
  over over swap 4* + scratch + -rot	\ scratch-idx r c
  4* + round.key +			\ scratch-idx rk-idx
  c@ over c@ xor swap c!
;

\ add round key
: ROUND-KEY+   ( -- )
  4 0 do
    4 0 do
      i j rk+calc
    loop
  loop
;

\ move input block to scratch
\ input data block (16-bytes)
: >AES   ( caddr -- )
  scratch swap
  4 0 do
    4 0 do				\ scratch data
      over over i + j 4* + swap		\ scratch data data-idx scratch
      i 4* + j + swap			\ scratch data scratch-idx data-idx
      c@ swap c!
    loop
  loop   2drop
;

\ move scratch to output block
\ output encrypted data block (16-bytes)
: AES>   ( c-addr -- )
  scratch
  4 0 do
    4 0 do				\ data scratch
      over over i 4* + j + swap		\ data scratch scratch-idx data
      i + j 4* + swap			\ data scratch data-idx scratch-idx
      c@ swap c!
    loop
  loop   2drop
;

\ store the first round key
: KEY-IN ( caddr -- )
  round.key aeslen cmove
;

\ perform one round of encryption
: ONE-ROUND ( round -- )
  s.b-all@
  sh-bytes
  dup 10 <> if mix-col then
  round-key
  round-key+
;

\ aes128 encrypt block
\ input block is ciphered in-situ
: AES   ( c-addr key -- )
  key-in
  dup >aes
  round-key+
  11 1 do
    i one-round
  loop
  aes>
;

\ alias
aka aes +AES
